# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Main module to run the Gordon Janitor service.

The service expects a ``gordon-janitor.toml`` and/or a
``gordon-janitor-user.toml`` file for configuration in the current
working directory, or in a provided root directory.

Any configuration defined in ``gordon-janitor-user.toml`` overwrites
those in ``gordon-janitor.toml``.

Example:

.. code-block:: bash

    $ python gordon_janitor/main.py
    $ python gordon_janitor/main.py -c /etc/default/
    $ python gordon_janitor/main.py --config-root /etc/default/
"""

import asyncio
import logging
import os

import click
import toml
import ulogger
from gordon import plugins_loader

from gordon_janitor import __version__ as version
from gordon_janitor import exceptions
from gordon_janitor import interfaces


plugins_loader.PLUGIN_NAMESPACE = 'gordon_janitor.plugins'


def _load_config(root=''):
    conf, error = {}, False
    conf_files = ['gordon-janitor.toml', 'gordon-janitor-user.toml']
    for conf_file in conf_files:
        try:
            with open(os.path.join(root, conf_file), 'r') as f:
                conf.update(toml.load(f))
        except IOError:
            error = True

    if error and conf == {}:
        raise IOError(f'Cannot load Gordon configuration file from "{root}".')
    return conf


def setup(config_root=''):
    """
    Service configuration and logging setup.

    Configuration defined in ``gordon-janitor-user.toml`` will overwrite
    ``gordon-janitor.toml``.

    Args:
        config_root (str): where configuration should load from,
            defaults to current working directory.
    Returns:
        A dict for Gordon service configuration
    """
    config = _load_config(root=config_root)

    logging_config = config.get('core', {}).get('logging', {})
    log_level = logging_config.get('level', 'INFO').upper()
    log_handlers = logging_config.get('handlers') or ['syslog']

    ulogger.setup_logging(
        progname='gordon-janitor', level=log_level, handlers=log_handlers)

    return config


def _log_or_exit_on_exceptions(base_msg, exc, debug):
    log_level_func = logging.warn
    if not debug:
        log_level_func = logging.error

    if isinstance(exc, list):
        for exception in exc:
            log_level_func(base_msg, exc_info=exception)
    else:
        log_level_func(base_msg, exc_info=exc)

    if not debug:
        raise SystemExit(1)


def _gather_providers(plugins, debug):
    # NOTE: this assumes dict ordering is deterministic, if ever ported
    #       to <3.6, this will break!
    providers = {
        'publisher': None,
        'reconciler': None,
        'authority': None,
    }
    for plugin in plugins:
        if interfaces.IPublisher.providedBy(plugin):
            providers['publisher'] = plugin
        elif interfaces.IReconciler.providedBy(plugin):
            providers['reconciler'] = plugin
        elif interfaces.IAuthority.providedBy(plugin):
            providers['authority'] = plugin

    missing = []
    msg = ('A provider for "{name}" interface is not configured for the '
           'Janitor service or is not implemented.')
    for provider, obj in providers.items():
        if obj is None:
            exc = exceptions.MissingPluginError(msg.format(name=provider))
            missing.append(exc)
    if missing:
        base_msg = 'Issue running plugins: '
        _log_or_exit_on_exceptions(base_msg, missing, debug=debug)

    return providers


async def _run(plugins, debug):
    # organize plugins to assert order; publisher should start first,
    # authority last
    providers = _gather_providers(plugins, debug=debug)

    tasks = []
    for name, provider in providers.items():
        try:
            tasks.append(provider.run())
        except AttributeError as e:
            base_msg = 'Plugin missing required "run" method: '
            _log_or_exit_on_exceptions(base_msg, name, debug=debug)

    await asyncio.gather(*tasks)


@click.command()
@click.option('-c', '--config-root',
              type=click.Path(exists=True), required=False, default='.',
              help='Directory where to find service configuration.')
def run(config_root):
    config = setup(os.path.abspath(config_root))
    debug_mode = config.get('core', {}).get('debug', False)

    # TODO: initialize a metrics object - either here or within `load_plugins`
    plugin_kwargs = {
        'rrset_channel': asyncio.Queue(),
        'changes_channel': asyncio.Queue(),
    }

    plugin_names, plugins, errors, plugin_kwargs = plugins_loader.load_plugins(
        config, plugin_kwargs)

    if errors:
        base_msg = 'Plugin was not loaded:'
        _log_or_exit_on_exceptions(base_msg, errors, debug_mode)

    if not plugin_names:
        logging.error('No plugins to run, exiting.')
        return SystemExit(1)

    logging.info(f'Loaded {len(plugin_names)} plugins: {plugin_names}')
    logging.info(f'Starting gordon janitor v{version}...')
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(_run(plugins, debug_mode))
    finally:
        loop.close()


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module for reusable pytest fixtures.
"""

import os

import pkg_resources
import pytest


class FakePlugin:
    def __init__(self, config, **kwargs):
        self.config = config

    async def run(self):
        pass

    async def cleanup(self):
        pass


@pytest.fixture
def plugin_exc_mock():
    class FakePluginException(Exception):
        """Exception raised from a plugin when loading"""
        pass
    return FakePluginException('dangit')


@pytest.fixture(scope='session')
def config_file():
    here = os.path.dirname(os.path.realpath(__file__))
    filepath = os.path.join(here, 'fixtures/test-gordon-janitor.toml')
    with open(filepath, 'r') as f:
        return f.read()


@pytest.fixture
def loaded_config():
    return {
        'core': {
            'plugins': [
                'xyz.authority',
                'xyz.reconciler',
                'xyz.publisher'],
            'debug': True,
            'logging': {
                'level': 'debug',
                'handlers': ['stream'],
                'format': '%(created)f %(levelno)d %(message)s',
                'date_format': '%Y-%m-%dT%H:%M:%S',
            }
        },
        'xyz': {
            'a_key': 'a_value',
            'b_key': 'b_value',
            'authority': {
                'a_key': 'another_value',
            },
            'reconciler': {
                'd_key': 'd_value',
            },
            'publisher': {
                'c_key': 'c_value',
            }
        }
    }


@pytest.fixture
def plugins(mocker):
    plugins = {}
    names = ['xyz.authority', 'xyz.reconciler', 'xyz.publisher']
    for name in names:
        plugin_mock = mocker.MagicMock(pkg_resources.EntryPoint, autospec=True)
        plugin_mock.name = name
        plugin_mock.load.return_value = FakePlugin
        plugins[name] = plugin_mock
    return plugins



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from click.testing import CliRunner

from gordon_janitor import interfaces
from gordon_janitor import main
from tests.unit import conftest


#####
# Tests for service setup
#####
@pytest.mark.parametrize('suffix', ['', '-user'])
def test_load_config(tmpdir, suffix, config_file, loaded_config):
    """Load prod and user config."""
    filename = 'gordon-janitor{}.toml'.format(suffix)
    conf_file = tmpdir.mkdir('config').join(filename)
    conf_file.write(config_file)
    config = main._load_config(root=conf_file.dirpath())

    assert loaded_config == config


def test_load_config_raises(tmpdir):
    """No config loaded raises IOError."""
    dir_with_no_conf = tmpdir.mkdir('config')
    with pytest.raises(IOError) as e:
        main._load_config(root=dir_with_no_conf.dirpath())

    assert e.match('Cannot load Gordon configuration file from')


def test_setup(tmpdir, mocker, monkeypatch, config_file, loaded_config):
    """Setup service config and logging."""
    conf_file = tmpdir.mkdir('config').join('gordon-janitor.toml')
    conf_file.write(config_file)

    ulogger_mock = mocker.MagicMock(main.ulogger, autospec=True)
    ulogger_mock.setup_logging = mocker.Mock()
    monkeypatch.setattr(main, 'ulogger', ulogger_mock)

    config = main.setup(config_root=conf_file.dirpath())

    assert loaded_config == config

    ulogger_mock.setup_logging.assert_called_once_with(
        progname='gordon-janitor', level='DEBUG', handlers=['stream'])


#####
# Tests & fixtures for running service
#####
@pytest.fixture
def setup_mock(mocker, monkeypatch):
    setup_mock = mocker.MagicMock(main.setup, autospec=True)
    monkeypatch.setattr(main, 'setup', setup_mock)
    return setup_mock


@pytest.fixture
def load_plugins_mock(mocker, monkeypatch):
    load_plugins_mock = mocker.MagicMock(
        main.plugins_loader.load_plugins, autospec=True)
    patch = 'gordon_janitor.main.plugins_loader.load_plugins'
    monkeypatch.setattr(patch, load_plugins_mock)
    return load_plugins_mock


args = 'error_type'
params = [
    'list', 'obj'
]


@pytest.mark.parametrize(args, params)
def test_log_or_exit_on_exceptions_no_debug(error_type, plugin_exc_mock,
                                            mocker, monkeypatch):
    """Raise SystemExit if debug flag is off."""
    logging_mock = mocker.MagicMock(main.logging, autospec=True)
    monkeypatch.setattr(main, 'logging', logging_mock)

    error = ('bad.plugin', plugin_exc_mock)
    if error_type == 'list':
        error = [('bad.plugin', error)]
    with pytest.raises(SystemExit) as e:
        main._log_or_exit_on_exceptions('base msg', error, debug=False)

    e.match('1')
    logging_mock.error.assert_called_once()
    logging_mock.warn.assert_not_called()


def test_log_or_exit_on_exceptions_debug(plugin_exc_mock, mocker, monkeypatch):
    """Do not exit out if debug flag is on."""
    logging_mock = mocker.MagicMock(main.logging, autospec=True)
    monkeypatch.setattr(main, 'logging', logging_mock)

    errors = [('bad.plugin', plugin_exc_mock)]

    main._log_or_exit_on_exceptions('base_msg', errors, debug=True)

    logging_mock.warn.assert_called_once()
    logging_mock.error.assert_not_called()


@pytest.fixture
def mock_provided_by(mocker, monkeypatch):
    mock_iauthority = mocker.Mock(interfaces.IAuthority, autospec=True)
    mock_iauthority.providedBy.side_effect = iter([True])
    patch = 'gordon_janitor.main.interfaces.IAuthority'
    monkeypatch.setattr(patch, mock_iauthority)

    mock_ireconciler = mocker.Mock(interfaces.IReconciler, autospec=True)
    mock_ireconciler.providedBy.side_effect = iter([True, False])
    patch = 'gordon_janitor.main.interfaces.IReconciler'
    monkeypatch.setattr(patch, mock_ireconciler)

    mock_ipublisher = mocker.Mock(interfaces.IPublisher, autospec=True)
    mock_ipublisher.providedBy.side_effect = iter([True, False, False])
    patch = 'gordon_janitor.main.interfaces.IPublisher'
    monkeypatch.setattr(patch, mock_ipublisher)

    return mock_iauthority, mock_ireconciler, mock_ipublisher


def test_gather_providers_no_providers(plugins, mock_provided_by, caplog):
    mock_iauthority, mock_ireconciler, mock_ipublisher = mock_provided_by
    mock_iauthority.providedBy.side_effect = [False, False, False]
    mock_ireconciler.providedBy.side_effect = [False, False, False]
    mock_ipublisher.providedBy.side_effect = [False, False, False]

    main._gather_providers(plugins, debug=True)

    assert 3 == len(caplog.records)


@pytest.mark.asyncio
async def test_run_no_providers(plugins, mock_provided_by, caplog):
    mock_iauthority, mock_ireconciler, mock_ipublisher = mock_provided_by
    mock_iauthority.providedBy.side_effect = [False, False, False]
    mock_ireconciler.providedBy.side_effect = [False, False, False]
    mock_ipublisher.providedBy.side_effect = [False, False, False]

    await main._run(plugins, debug=True)

    assert 6 == len(caplog.records)


@pytest.mark.asyncio
@pytest.mark.parametrize('debug', (True, False))
async def test_async_run_debug(debug, mock_provided_by, caplog):
    plugins = [
        conftest.FakePlugin({}),
        conftest.FakePlugin({}),
        conftest.FakePlugin({})
    ]
    await main._run(plugins, debug=debug)

    mock_iauthority, mock_ireconciler, mock_ipublisher = mock_provided_by
    assert 3 == mock_ipublisher.providedBy.call_count
    assert 2 == mock_ireconciler.providedBy.call_count
    assert 1 == mock_iauthority.providedBy.call_count


run_args = 'has_active_plugins,exp_log_count'
run_params = [
    (True, 2),
    (False, 1),
]


@pytest.mark.parametrize(run_args, run_params)
def test_run(has_active_plugins, exp_log_count, plugins, setup_mock,
             load_plugins_mock, mock_provided_by, mocker, monkeypatch, caplog):
    """Successfully start the Gordon service."""
    names, _plugins, errors, kwargs = [], [], [], {}
    if has_active_plugins:
        names = ['authority.plugin', 'reconciler.plugin', 'publisher.plugin']
        _plugins = [
            conftest.FakePlugin({}),
            conftest.FakePlugin({}),
            conftest.FakePlugin({})
        ]
    load_plugins_mock.return_value = names, _plugins, errors, kwargs

    runner = CliRunner()
    result = runner.invoke(main.run)

    assert 0 == result.exit_code
    setup_mock.assert_called_once()
    assert exp_log_count == len(caplog.records)


def test_run_raise_exceptions(loaded_config, plugins, caplog, setup_mock,
                              load_plugins_mock, plugin_exc_mock,
                              mock_provided_by, monkeypatch, mocker):
    """Raise plugin exceptions when not in debug mode."""
    loaded_config['core']['debug'] = False
    setup_mock.return_value = loaded_config

    names = ['authority.plugin', 'reconciler.plugin']
    errors = [('publisher.plugin', plugin_exc_mock)]
    kwargs = {}
    load_plugins_mock.return_value = names, plugins, errors, kwargs

    runner = CliRunner()
    result = runner.invoke(main.run)

    assert 1 == result.exit_code
    setup_mock.assert_called_once()
    assert 1 == len(caplog.records)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import os
import re

from setuptools import find_packages
from setuptools import setup


HERE = os.path.abspath(os.path.dirname(__file__))


#####
# Helper functions
#####
def read(*filenames, **kwargs):
    """
    Build an absolute path from ``*filenames``, and  return contents of
    resulting file.  Defaults to UTF-8 encoding.
    """
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for fl in filenames:
        with codecs.open(os.path.join(HERE, fl), 'rb', encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


def find_meta(meta):
    """Extract __*meta*__ from META_FILE."""
    re_str = r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta)
    meta_match = re.search(re_str, META_FILE, re.M)
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError('Unable to find __{meta}__ string.'.format(meta=meta))


def install_requires():
    reqs_txt = read('requirements.txt')
    parsed = reqs_txt.split('\n')
    parsed = [r.split('==')[0] for r in parsed]
    return [r for r in parsed if len(r) > 0]


#####
# Project-specific constants
#####
NAME = 'gordon-janitor'
PACKAGE_NAME = 'gordon_janitor'
PACKAGES = find_packages(where='.')
META_PATH = os.path.join(PACKAGE_NAME, '__init__.py')
KEYWORDS = ['dns']
CLASSIFIERS = [
    'Development Status :: 1 - Planning',
    'License :: OSI Approved :: Apache Software License',
    'Natural Language :: English',
    'Operating System :: POSIX :: Linux',
    'Operating System :: MacOS :: MacOS X',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: Implementation :: CPython',
    'Topic :: Internet :: Name Service (DNS)',
]
META_FILE = read(META_PATH)


setup(
    name=NAME,
    version=find_meta('version'),
    license=find_meta('license'),
    description=find_meta('description'),
    url=find_meta('uri'),
    author=find_meta('author'),
    author_email=find_meta('email'),
    maintainer=find_meta('author'),
    maintainer_email=find_meta('email'),
    packages=PACKAGES,
    entry_points={
        'console_scripts': [
            'gordon-janitor = gordon_janitor.main:run'
        ],
        'gordon_janitor.plugins': []
    },
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    zip_safe=False,
    install_requires=install_requires(),
)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface definitions for Gordon Janitor Plugins.

Please see :doc:`plugins` for more information on writing a plugin for
the Gordon Janitor service.
"""

from zope.interface import Interface


class IGenericPlugin(Interface):
    """**Do not** implement this interface directly.

    Use :py:class:`IAuthority`, :py:class:`IReconciler`, or
    :py:class:`IPublisher` instead.

    Args:
        config (dict): Plugin-specific configuration.
        plugin_kwargs (dict): Plugin-specific keyword arguments. See
            specific interface declarations.
    """
    def __init__(config, **plugin_kwargs):
        """Initialize a Janitor Plugin client."""

    async def run():
        """Start plugin in the main event loop.

        Once required work is all processed, :py:meth:`cleanup` needs to
        be called.
        """

    async def cleanup():
        """Cleanup once plugin-specific work is cleanup.

        Cleanup work might include allowing outstanding asynchronous
        Python tasks to finish, cancelling them if they extend beyond a
        desired timeout, and/or closing HTTP sessions.
        """


class IAuthority(IGenericPlugin):
    """Scan source of truth(s) of hosts and emit messages to Reconciler.

    The purpose of this client is to consult a source of truth, for
    example, the list instances APIs in Google Compute Engine or AWS
    EC2, or consulting one's own database of hosts. A message per DNS
    zone with every instance record (per service owner's own
    requirements) will then be put onto the ``rrset_channel`` queue for
    a Reconciler to - you guessed it - reconcile.

    Args:
        config (dict): Authority-specific configuration.
        rrset_channel (asyncio.Queue): queue to put record set messages
            for later validation.
        metrics (obj): Optional object to emit Authority-specific
            metrics.
    """
    def __init__(config, rrset_channel, metrics=None):
        """Initialize an Authority Plugin client."""


class IReconciler(IGenericPlugin):
    """Validate current records in DNS against desired Authority.

    Clients that implement :py:class:`IReconciler` will create a change
    message for the configured :py:class:`IPublisher` client plugin to
    consume if there is a discrepancy between records in DNS and the
    desired state.

    Once validation is done, the :py:class:`IReconciler` client will
    need to emit a ``None`` message to the :py:attr:`changes_channel`
    queue, signalling to an :py:class:`IPublisher` client to publish the
    message to a pub/sub to which `Gordon
    <https://github.com/spotify/gordon>`_ subscribes.

    Args:
        config (dict): Reconciler-specific configuration.
        rrset_channel (asyncio.Queue): queue from which to consume
            record set messages to validate.
        changes_channel (asyncio.Queue): queue to publish corrective
            messages to be published.
        metrics (obj): Optional object to emit Reconciler-specific
            metrics.
    """
    def __init__(config, rrset_channel, changes_channel, metrics=None):
        """Initialize a Reconciler Plugin client."""


class IPublisher(IGenericPlugin):
    """Publish change messages to the pub/sub Gordon consumes.

    Clients that implement :py:class:`IPublisher` will consume from the
    :py:attr:`changes_channel` queue and publish the message to the
    configured pub/sub for which `Gordon
    <https://github.com/spotify/gordon>`_ subscribes.

    Args:
        config (dict): Publisher-specific configuration.
        changes_channel (asyncio.Queue): queue to consume the
             corrective messages needing to be published.
        metrics (obj): Optional object to emit Publisher-specific
            metrics.
    """
    def __init__(config, changes_channel, metrics=None):
        """Initialize a Publisher Plugin client."""
